
from bj_nlp            import blpClass
from spacy.tokens      import Doc, Token  #, Span
from spacy.matcher     import Matcher
#from bpyth             import load_pickle


# Funktionen tagZ, tagZZ
from bj_nlp.translate_tags  import *





class Merger:
    '''
    Pipeline-Komponente
    merger
    ==========
    Sucht nach Abkürzungen und fasst diese zu einem Token zusammen.
    Die Suche geschieht über RexEx sowie durch Nachschlagen im Wiktionary-Wortschatz.
    Diese Komponente muss als allererstes in der Pipeline liegen, damit die gemergten Token verarbeitet werden.
    
    Registrierung siehe blpClass.
    
    # Aktivierung:
    if not 'merger' in nlp.pipe_names:
        nlp.add_pipe('merger', 
                     first=True,          
                     config={'debug': True} 
                     ) 
        
    # direkter Zugriff auf die Pipeline-Komponente:
    merger = blp.nlp.get_pipe('merger') 
    
    .
    '''    
    
    # ===========================================================================================0 
    # init
    #    
    
    def __init__(self, blp: blpClass, debug: bool):
        
        
        self.blp   = blp                    # das blp-Objekt
        self.name  = 'merger'           # name, für Error-Messages 
        self.debug = debug
                    
        if debug:  

            if not Token.has_extension('debug'):
                Token.set_extension(   'debug', default='')               

        # load lookups
        try:
            import importlib.resources as pkg_resources
        except ImportError:
            # Try backported to PY<37 `importlib_resources`.
            import importlib_resources as pkg_resources
        from . import data     

        import pickle
        text = pkg_resources.read_binary( data, 'lookup1.pickle' )
        self.lookup1 = pickle.loads(text)      
        text = pkg_resources.read_binary( data, 'lookup4.pickle' )
        self.lookup4 = pickle.loads(text)      
            
            
        # === Matcher zur Erkennung von Abkürzungen
        self.matcher = Matcher(blp.nlp.vocab)
        patterns  = []
        
        # Ein oder zwei Zeichen, dann Punkt, dann wieder ein oder zwei Zeichen 
        token1 = {"TEXT": {"REGEX": "^.{1,2}\..{1,2}$"}}
        token2 = {"ORTH": "."}
        patterns +=  [[  token1, token2 ]] 
        
        # Ein oder zwei Konsonanten
        wird_nicht_mit_punkt_abgekürzt = ['au','ei','hm','qm','ui','mg','km','cm','mm','dm','ft','nm',
                                          'pt','m²','m³','kg','lb','db','BH','CD','CC','CT','DB','DC','DM','FH','KW','VW','XY']
        token1 = {"TEXT": {"REGEX": "^[b-df-hj-np-tv-zßB-DF-HJ-NP-TV-Z]{1,2}$"},  
                  "ORTH": {"NOT_IN": wird_nicht_mit_punkt_abgekürzt}
                 }           
        token2 = {"ORTH": "."}
        patterns +=  [[  token1, token2 ]]          
        
        #add patterns
        self.matcher.add('abbrev', patterns)  
               
        # fertig
        if self.blp.verbose:
            print( self.name + ' initialisiert')           
        


    
    
    # ===========================================================================================0
    # Hauptroutine 
    #    
    def __call__(self, doc:Doc) -> Doc:
    
        
        if len(doc) <= 1:
            return doc  
        

        
        # --------------------------------------------------------
        # Turn 0: Abkürzungen per Matcher suchen
        #               
            
        matches = self.matcher(doc)
        spans = []  # Collect the matched spans here
        for match_id, start, end in matches:
            spans.append(doc[start:end])
            if self.debug:
                try:
                    doc[start]._.debug += ' ↰'
                except:
                    pass
            try:
                doc[end].is_sent_start = False
            except:
                pass                
        
        # Token zusammenfassen
        with doc.retokenize() as retokenizer:
            for span in spans:
                retokenizer.merge(span)          
        
        
        
        # --------------------------------------------------------
        # Turn 1: Retokenize laut Wörterbuch
        # versucht zwei Token zu einem zusammenzufassen
        # wenn zwischen den Token ein Space ist, bleibt dieses erhalten!
        
        with doc.retokenize() as retokenizer:
            
            verschmolzen    = ''  
            verschmolzen_id = -1
            
            # Doppelwörter suchen 
            for i,token in enumerate(doc):
                    
                if i >= len(doc)-1:
                    continue
                    
                if doc[i].tag_ == '_SP':
                    continue
                    
                if doc[i+1].tag_ == '_SP':
                    continue                                    
                    
                span      = doc[i:i+2]  
                span_hash = self.blp.str2hash( span.text.lower() )
                eintrag   = self.lookup1.get( span_hash, None )      
                
                # Zweiter versuch
                if not eintrag: 
                    eintrag   = self.lookup4.get( span_hash, None ) 
                    
                # Immer noch nichts gefunden >> weiter                      
                if not eintrag: 
                    continue                    
                    
                # gefunden >> zu einem Token verschmelzen                                       
                # für Errors merken    
                verschmolzen_id_pre = verschmolzen_id
                verschmolzen_id     = i
                verschmolzen_pre    = verschmolzen    
                verschmolzen        = span                    

                if self.debug:
                    doc[i]._.debug     += ' ⨝'      
                    
                # Abkürzung?
                if span.text[-1] == '.':
                    try:                      
                        doc[i+2].is_sent_start = False
                    except:
                        pass                      
                    
                      
                # verschmelzen
                try: 
                    retokenizer.merge(span) 
                except ValueError:         
                    error = {'error'    : 'Überlappende Verschmelzungen',
                            'component' : self.name, 
                             'nam0'     : 'verschmolzen_pre',
                             'val0'     : verschmolzen_pre,  
                             'nam1'     : 'verschmolzen',
                             'val1'     : verschmolzen,      
                             #'nam2'     : 'doc_id',
                             #'val2'     : self.spacy_batch.doc_id,          
                             'nam3'     : 'pos',
                             'val3'     : i                                
                            }
                    if self.debug:
                        doc[i]._.debug += str(error)     
                        
                    
        
        # retokenizer end
 
                
        return doc
                
                
 
    
    
    
    
    
    
    
    
    
    
    # ===========================================================================================0 
    # Tests
    #
    
    
    TESTTEXT_0 = '''Tägl. kauft Mrs. Summer neue CDs.
Hallo Hn.g. so ist das.'''
        
    zweierworte =  ['ab','am','an','aß','da','du','eh','er','es','he','im','in','ja','je','la','mg',]
    zweierworte += ['ne','ob','oh','on','so','tu','um','wo','zu','äh','au','bi','ei','ex','ey','go','hi',]
    zweierworte += ['hm','ho','hä','hü','is','me','mi','na','nu','nö','or','qm','re','si','sä',]
    zweierworte += ['to','uh','ui','up','uz','wa','we','öd','öl','ös','üb','μm']
    zweierworte = [z+'.' for z in zweierworte]
    TESTTEXT_1 = ' '.join(zweierworte) #+ ' ' + ''.join(zweierworte)       
    
    # soll nicht verbunden werden
    TESTTEXT_2 = '''Schleswig Holstein San Franzisko'''    


    
    TESTTEXTE = { 0:TESTTEXT_0, 1:TESTTEXT_1, 2:TESTTEXT_2 }
    

    
    
    # ===========================================================================================0 
    # tests durchführen
    #
    def tests(self, only=None, verbose=True):           
        
        # Test 0: doc erstellen
        text = self.TESTTEXTE[0]
        sent_anz = text.count('\n')+1
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz
        
        # sent und sent_pre
        assert doc[2]._.sent   == 'Mrs. Summer kauft eine CD.'
        assert doc[7]._.sent   == 'Fast tägl. kauft Mrs. Summer neue CDs!'          
        assert doc[17]._.sent  == 'Tägl. kauft Mrs. Summer neue CDs.'   
        assert doc[23]._.sent  == 'Hallo Hn.g. so ist das.'
        
        # tagZ und tagZZ
        assert doc[2]._.tagZ == 'VER'
        assert doc[2]._.tagZZ == 'V'    
        
        # lookup1    
        test = self.lookup1[  7608359307979301264  ]    
        assert test[0] == 'heimtrauen'        
        
        if verbose:
            print('wortschatz Tests 0 OK')              
        
        
        # Test1: doc erstellen
        text = self.TESTTEXTE[1]
        sent_anz = text.count('\n')+1        
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz        
        
        # lemmaW
        assert doc[0]._.lemmaW_ == 'Lena'
        assert doc[1]._.lemmaW_ == ''  
        assert doc[2]._.lemmaW_ == 'laufen'          
        
        if verbose:
            print('wortschatz Tests 1 OK')          
            
            
        # Test2: doc erstellen
        text = self.TESTTEXTE[2]
        sent_anz = text.count('\n')+1        
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz  
        
        if verbose:
            print('wortschatz Tests 2 OK')   
            
            
        # Test3: doc erstellen
        text = self.TESTTEXTE[3]
        sent_anz = text.count('.')       
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz          
          
        if verbose:
            print('wortschatz Tests 3 OK')    
            
            
        # Test4: Mehr oder weniger leeres Dokument
        texte = self.sent_start + self.sent_end + ['']
        for text in texte:
            #print(text)
            doc = self.blp.nlp(text)
            assert len(doc) == len(text)     
            
            doc = self.blp.nlp(text + '.')
            assert len(doc) <= len(text) + 1     
            
            doc = self.blp.nlp('»' + text)
            assert len(doc) == len(text) + 1            
            
        
        if verbose:
            print('wortschatz Tests 4 OK')           
            
        return True
      
            
            
#          # Wortschatz prüfen    
#          prüfe_tagZ( 'X_konj',        ['z.B.','z. B.','Z.B.', 'D.h.','D. h.','d.h.', 'd. h.']   , doc )
#          prüfe_tagZ( 'ADV',           ['U.a.','u.a.','u. a.']   , doc )            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            




# ===========================================================================================0
# TESTTEXTE
#          

# Preprocess
TESTTEXT_0 = '''Hällö, „Welt“! <i>Hel 9lo</i> <b>Wo9 rld</b>! Das Wetter ist gut. Wie könnte`s not? A Öder nicht? Nic nemůže letět Rychlostí vyšší, než 300 tisíc kilometrů za sekundu! Hel 9lo Wo9rld! Th3     weather_is really g00d today, isn't it? Das geht    super, nöß?   Riß_2'''


# Kokosnuss
TESTTEXT_DE1 = '''»Dies ist Björns Buch.«
In einer Bucht unterhalb der schönsten Höhle gehen sie vor Anker.
Nachdem sie Feuerholz gesammelt haben, angeln sie für das Abendessen ein paar Fische.
»Ziemlich kalt hier«
, sagte Matilda und fröstelt.
»Dort hinten«
, sagt Oskar,
»habe ich vorhin ein Eisbärfell gesehen.
Wenn du das trocknest, hast du eine schöne warme Decke.«
Kokosnuss stutzt.
»Ein Eisbärfell?
Das muss ich sehen!'''  



TESTTEXT_EN1 = '''Apple is looking at buying U.K. startup for $1 billion.
I was reading the paper.'''







    
#import pandas     as pd 
import pandasklar as pak 


# key         0                     1            2          3         4  
# TAG         tag_help              tagZ         is_mutable to_lemma  tagZZ
bj_tags = {

  'AFIX':   ['Präfix',              'FIX',       False,     'W',       'Z'    ],    # Fixe disqualifizieren sich meistens, wenn man N, V oder A sucht 
  'MFIX':   ['Midfix',              'FIX',       False,     'W',       'Z'    ],         
  'ZFIX':   ['Postfix',             'FIX',       False,     'W',       'Z'    ],     

  'NN'  :   ['Nomen',               'NOM',       True,      'N',       'N'    ],
  'NE'  :   ['Eigenname',           'NAM',       True,      'W',       'N'    ],  
  'NNE' :   ['Eigenname',           'NAM',       True,      'W',       'N'    ],       

  'ADV':    ['Adverb',              'ADV',       False,     'W',       'A'    ],    # links, hier, heute, vielleicht, deshalb 
  'PROAV':  ['Adverb',              'ADV',       False,     'W',       'A'    ],     
  'PTKA':   ['Adverb',              'ADV',       False,     'W',       'A'    ],    # zu am     

  'ADJA':   ['Adjektiv',            'ADJ',       True,      'A',       'A'    ],    # gut klein ganz rosa
  'ADJD':   ['Adjektiv',            'ADJ',       True,      'A',       'A'    ], 
  'VVPP':   ['Ptz2',                'ADJ',       True,      'V',       'A'    ],    # ausprobiert   (ist wie ein Adjektiv)     

  'VVIZU':  ['Verb',                'VER',       True,      'V',       'V'    ],  

  'VVFIN':  ['Verb',                'VER',       True,      'V',       'V'    ],    # fragen sagen sehen   
  'VVIMP':  ['Verb',                'VER',       True,      'V',       'V'    ],       
  'VVINF':  ['Verb',                'VER',       True,      'V',       'V'    ],  
  'PTKVZ':  ['Verb',                'VER',       False,     'V',       'V'    ],    # inne, dar, zurecht

  'VMFIN':  ['Modalverb',           'VER_mod',   True,      'V',       'V'    ],    # müssen sollen können
  'VMINF':  ['Modalverb',           'VER_mod',   True,      'V',       'V'    ],  
  'VMPP':   ['Modalverb',           'VER_mod',   True,      'V',       'V'    ],       
  'VAFIN':  ['Hilfsverb',           'VER_hilf',  True,      'V',       'V'    ],    # sein haben 
  'VAIMP':  ['Hilfsverb',           'VER_hilf',  True,      'V',       'V'    ],   
  'VAINF':  ['Hilfsverb',           'VER_hilf',  True,      'V',       'V'    ],  
  'VAPP':   ['Hilfsverb',           'VER_hilf',  True,      'V',       'V'    ],  


  'PPER':   ['Personalpronomen',    'PRO',       True,      'W',       'P'    ],    # ich mir er 
  'PRF':    ['Personalpronomen',    'PRO',       True,      'W',       'P'    ],    
  'PPOSAT': ['Possessivpronomen',   'PRO',       True,      'W',       'P'    ],    # mein dein sein         
  'PPOSS':  ['Possessivpronomen',   'PRO',       True,      'W',       'P'    ],    
  'PRELAT': ['Relativpronomen',     'PRO',       True,      'W',       'P'    ],    # mich mir                              # Kann ev. auch lemmatisiert werden
  'PRELS':  ['Reflexivpronomen',    'PRO',       True,      'W',       'P'    ],     
  'PDAT':   ['Demonstrativpronomen','PRO',       True,      'W',       'P'    ],    # dieser jener der   
  'PDS':    ['Demonstrativpronomen','PRO',       True,      'W',       'P'    ],    
  'PIAT':   ['Indefinitpronomen',   'PRO',       True,      'W',       'P'    ],    # manche alle keiner  
  'PIS':    ['Indefinitpronomen',   'PRO',       True,      'W',       'P'    ],       
  'PWAT':   ['Fragepronomen',       'PRO',       True,      'W',       'P'    ],    # wer was wem wessen welcher   
  'PWAV':   ['Fragepronomen',       'PRO',       True,      'W',       'P'    ],    
  'PWS':    ['Fragepronomen',       'PRO',       True,      'W',       'P'    ],     
  'PRO':    ['Pronomen',            'PRO',       True,      'W',       'P'    ],     


  'APPO':   ['Präposition',         'X_prep',    False,     'W',       'X'    ],    # im auf
  'APPR':   ['Präposition',         'X_prep',    False,     'W',       'X'    ],     
  'APPRART':['Präposition',         'X_prep',    False,     'W',       'X'    ], 
  'APZR':   ['Präposition',         'X_prep',    False,     'W',       'X'    ],     
  'KOKOM':  ['Konjunktion',         'X_konj',    False,     'W',       'X'    ],    # und aber als
  'KON':    ['Konjunktion',         'X_konj',    False,     'W',       'X'    ],       
  'KOUI':   ['Konjunktion',         'X_konj',    False,     'W',       'X'    ], 
  'KOUS':   ['Konjunktion',         'X_konj',    False,     'W',       'X'    ],    
  'ART' :   ['Artikel',             'X_art',     True,      'W',       'X'    ],    # der die das ein eine     
  'PTKANT': ['Antwortpartikel',     'X_antw',    False,     'W',       'X'    ],    # ja nein bitte danke
  'PTKNEG': ['Negationspartikel',   'X_neg',     False,     'W',       'X'    ],    # nicht      
  'CARD':   ['Zahl',                'X_zahl',    False,     'W',       'X'    ],    # eins zwei 3
  'ITJ':    ['Interjektion',        'X_itj',     False,     'W',       'X'    ],    # Hallo ach  
  'KOMBI':  ['Wortkombination',     'X_kombi',   False,     'W',       'X'    ],    # Dampf ablassen, Seine Heiligkeit
  'DIV':    ['Diverses',            'X_div',     False,     'W',       'X'    ],      

  '$('  :   ['Satzzeichen',         'SYM',       False,     'W',       'Y'    ],    
  '$,'  :   ['Satzzeichen',         'SYM',       False,     'W',       'Y'    ],  
  '$.'  :   ['Satzzeichen',         'SYM',       False,     'W',       'Y'    ],   
  '_SP':    ['Satzzeichen',         'SYM',       False,     'W',       'Y'    ],          

  'PTKZU':  ['Müll',                'z',         False,     'W',       'Z'    ],    # zu   
  'XY':     ['Müll',                'z',         False,     'W',       'Z'    ],     
  'TRUNC':  ['Müll',                'z',         False,     'W',       'Z'    ],     
  'FM':     ['Fremdsprachliches',   'z',         False,     'W',       'Z'    ],    

  'TODO':   ['(todo!)',             'X_todo',    False,     'W',       'X'    ],       

}

    
# erstellt ein DataFrame namens translate_tagZ
# zur Konvertierung tag >> tagZ >> tagZZ (verschiedene Abstraktionsstufen der Tags)
# sowie Wortarten wie sie im Wiktionary verwendet werden zu tags 
#
def generate_translate_tags():    
        
    # in DataFrame wandeln   
    result = pak.dataframe(bj_tags, verbose=False).transpose().reset_index()
    result.columns = ['tag','wortart_lang','tagZ','is_mutable','to_lemma','tagZZ']
    result = pak.move_cols(result, 'tagZZ','tagZ')
    return result




# interne Funktion für den Zugriff auf bj_tags
def bj_key( key, position ):    
    if key in bj_tags:
        return bj_tags[key][position]
    else:
        if position == 0:
            return spacy.explain(key) 
        elif position == 1:
            return '¿¿¿'     # tagZ
        elif position == 2:
            return False     # is_mutable  
        elif position == 3:
            return 'W'       # to_lemma  
        elif position == 4:
            return '¿'       # tagZZ           
        

# Langform von tag
def tag_help(key):
    return bj_key(key,0)        
        
    

def tagZ( tag ):
    '''Liefert tagZ, ein zusammenfassendes tag.
    Dies ist die Funktion bj_nlp.tagZ, die Strings annimmt.
    Es gibt auch die Methode blp.tagZ des blp-Objektes, die Hashwerte annimmt.
    '''
    return bj_key(tag,1)         

        

def tagZZ( tag ):
    '''Liefert tagZZ, ein stark zusammenfassendes tag.
    Dies ist die Funktion bj_nlp.tagZZ, die Strings annimmt.
    Es gibt auch die Methode blp.tagZZ des blp-Objektes, die Hashwerte annimmt.
    '''    
    return bj_key(tag,4)   



# Wird das Wort konjugiert, dekliniert o.ä.?
# Kann TEXT!=LEMMA ?   
def is_mutable(key):
    return bj_key(key,2) 


# Hilfsfunktion für Lemma2
# siehe Lemma2
#
def to_lemma(key):
    return bj_key(key,3) 




# Liefert eine Liste von  Wortarten
def wortarten(key='alle'):
    alle_tags = list(bj_tags.keys())
    alle_tags.sort()
    if key == 'alle':
        return alle_tags 
    else:
        result = [tag for tag in alle_tags if wortart(tag).startswith(key)  ]
        return result 
    




##################################################################################################
    
bj_tagZ = {
    'NOM'     :  'NN',
    'NAM'     :  'NE',
    'ADV'     :  'ADV',    
    'ADJ'     :  'ADJA',       
    'VER'     :  'VVFIN',       
    'VER_mod' :  'VMFIN',   
    'VER_hilf':  'VAFIN',     
    'PRO'     :  'PDAT',
    'X_prep'  :  'APPRART',    
    'X_art'   :  'ART',     
    'X_antw'  :  'PTKANT',  
    'X_neg'   :  'PTKNEG',  
    'X_zahl'  :  'CARD',  
    'X_itj'   :  'ITJ',    
    '$('      :  'SYM',     
    'z'       :  '',       
}

# Liefert einen passenden tag, wenn nur tagZ gegeben ist
def tag(key):
    if key in bj_tagZ:
        return bj_tagZ[key]
    return ''




# ===========================================================================================0
# Test-Utils
#    
    
    
# Generator, liefert alle token mit einem bestimmten text     
def find_token(text, doc):
    for token in doc:
        if token.text == text:
            yield token     




def check_tagZ(soll_sein, check_token, doc):
    '''
    # Beispiel    
    bj_nlp.check_tagZ( 'X_konj',        ['z.B.','z. B.','Z.B.', 'D.h.','D. h.','d.h.', 'd. h.']   , doc )
    bj_nlp.check_tagZ( 'ADV',           ['U.a.','u.a.','u. a.']   , doc )    
    '''
    for check in check_token:
        tokenliste = list(find_token(check, doc))
        assert len(tokenliste) > 0
        for token in tokenliste:
            tagz = tagZ(token.tag_)
            if tagz != soll_sein:
                print(token.text + ':', 'ist:' + tagz, 'soll:'+soll_sein)
            assert tagz == soll_sein       


from spacy.language    import Language
    
from bj_nlp.translate_tags    import *
from bj_nlp._PandaziseMixin   import *    # Schnittstellen zu Pandas
from bj_nlp._UtilsMixin       import *    # Wichtige Klassenmethoden


try:
    from dframcy import DframCy
except:
    pass

# blpClass ist die Klasse des blp-Objektes, das einmal instanziiert wird.
# PandaziseMixin und UtilsMixin werden hineingemixt.
#
class blpClass( PandaziseMixin, UtilsMixin ):
    
# ===========================================================================================0
# __init__
#       
    
    def __init__( self, lang='en', model=0, verbose=True ):  
        '''
        Initialisierung z.B.:
        if not 'blp' in globals() or True:
            blp = bj_nlp.blpClass(lang='de')        
            
        Statt nlp verwendet man dann blp.nlp
        
        
        '''
        
        # merken
        self.lang    = lang
        self.model   = model
        self.verbose = verbose
        
        # geladenes model
        self.model_name = 'blank'        
        
        # translate_tagZ (DataFrame zum Übersetzen tag > tagZ > tagZZ)
        self.translate_tagZ = generate_translate_tags()        
        
        
        # nlp-Instanz / load model        
        if lang == 'en':
            if model == 0:
                self.model_name = 'en_core_web_sm' 
            elif model == 1:           
                self.model_name = 'en_core_web_trf'  
            elif model == 2:           
                self.model_name = 'en_core_web_md'  
            elif model == 3:           
                self.model_name = 'en_core_web_lg'                  
                
        elif lang == 'de':     
            if model == 0:
                self.model_name = 'de_core_news_sm' 
            elif model == 1:           
                self.model_name = 'de_core_news_trf'  
            elif model == 2:           
                self.model_name = 'de_core_news_md'  
            elif model == 3:           
                self.model_name = 'de_core_news_lg'       

                
        if self.model_name != 'blank': 
            self.nlp = spacy.load(self.model_name)
        else:
            self.nlp = spacy.blank(lang)
            
        
        if self.verbose:
            print(self.model_name, 'loaded')       
            
        # vocab korrigieren
        lex = self.nlp.vocab['»']
        lex.is_left_punct = True
        lex.is_right_punct = False

        lex = self.nlp.vocab['«']
        lex.is_left_punct = False
        lex.is_right_punct = True                
            

        # DframCy-Instanz
        try:
            self.dframcy = DframCy(self.nlp)        
        except:
            pass
        
        
        # DE-Pipeline-Komponenten registrieren
        if lang=='de':
            
            # === merger ===             
            if not Language.has_factory('merger'):
                if self.verbose:
                    print('registriere merger')
                @Language.factory('merger',
                                  assigns=[], 
                                  retokenizes=True,
                                  default_config={'debug': True}
                                 )
                def create_wortschatz_component(nlp: Language, name: str, debug: bool):
                    return bj_nlp.Merger(nlp.blp, debug)    
            else:
                if self.verbose:
                    print('merger bereits registriert')              
                                
                        
                        
            # === wrede ====
            if not Language.has_factory('wrede'):
                if self.verbose:
                    print('registriere wrede')
                @Language.factory('wrede', 
                                  assigns=['token._.is_wrede','doc._.wrede'],
                                  default_config={'debug': True}
                                 )
                def create_wrede_component(nlp: Language, name: str, debug: bool):
                    return bj_nlp.WRede(nlp.blp, debug)    
            else:
                if self.verbose:
                    print('wrede bereits registriert')   
                    
                    

                    
            # === wortschatz ===             
            if not Language.has_factory('wortschatz'):
                if self.verbose:
                    print('registriere wortschatz')
                @Language.factory('wortschatz',
                                  assigns=['token._.tagZ',   'token._.tagZZ',
                                           'token._.sent',   'token._.sent_pre', 'token._.is_sent_start',
                                           'token._.lemmaW', 'token._.lemmaW_',], 
                                  retokenizes=True,
                                  default_config={'debug': True}
                                 )
                def create_wortschatz_component(nlp: Language, name: str, debug: bool):
                    return bj_nlp.Wortschatz(nlp.blp, debug)    
            else:
                if self.verbose:
                    print('wortschatz bereits registriert')                      

  
                    
                   
            
            
        # TESTTEXTE
        #self.TESTTEXTE = bj_nlp.TESTTEXTE
        
        if lang=='de':        
            self.TESTTEXTE = {       0  :  bj_nlp.TESTTEXT_0, 
                                     1  :  bj_nlp.TESTTEXT_DE1,
                          'Preprocess'  :  bj_nlp.TESTTEXT_0,               
                           'Kokosnuss'  :  bj_nlp.TESTTEXT_DE1,

                        }   
        else:        
            self.TESTTEXTE = {       0  :  bj_nlp.TESTTEXT_0, 
                                     1  :  bj_nlp.TESTTEXT_EN1,                              
                          'Preprocess'  :  bj_nlp.TESTTEXT_0,               

                        }                
    
    




#####################################################################################################
# Schnittstellen zu den Klassenmethoden
#####################################################################################################  

import bj_nlp
import bpyth as bpy

from spacy.matcher import PhraseMatcher

class UtilsMixin:
    
#####################################################################################################
# Schnittstellen zu den Klassenmethoden
#####################################################################################################  
    
    
    def testtext(self, variant='Kokosnuss'):
        return bj_nlp.testtext(variant=variant)
    
    
    def preprocess(self, text, config=0):
        return bj_nlp.preprocess(text, config=config)
    
    
    

#####################################################################################################
# Eigene Utils
#####################################################################################################  
    


    
    def phrasematcher( self, terms=[], match_id='no_id', on_match=None, attr='ORTH', validate=False, verbose=False):
        '''
        
        Liefert einen einfachen PhraseMatcher mit den angegebenen Rules.
        
        terms kann eine Liste von Strings oder eine Liste von Dictionaries sein.
        
        === list of dictionaries ===
        terms ist eine list of dictionaries mit den keys 'match_id' und 'text', z.B.:
        [{  'match_id': 'WREDE_L',   'text': '»'  },
         {  'match_id': 'WREDE_R',   'text': '«'  }, ]
         
        Oder, wenn eine Callback-Funktion angegeben werden soll:
        [{  'match_id': 'WREDE_L',   'text': '»',   'on_match': wrede_func  },
         {  'match_id': 'WREDE_R',   'text': '«',   'on_match': wrede_func   }, ] 
        Eine Callback-Funktion pro match_id ist möglich. 
        Alternativ kann die Callback-Funktion auch global über den Parameter on_match angegeben werden.
        
        === list of strings ===
        terms kann auch einfach eine Liste von Strings sein. 
        Dann können match_id und on_match
        '''
        
        #print(bpy.rtype(terms))

        matcher = PhraseMatcher(self.nlp.vocab, attr=attr, validate=validate)
        rtype_terms = bpy.rtype(terms)
        
        # terms ist Liste von Dictionaries 
        if rtype_terms[:2] == ('list', 'dict' ):
            
            # docs erstellen
            for term in terms:
                term['doc'] = self.nlp.make_doc( term['text']  )       

            # für jede id eine Rule
            matchids = { term['match_id'] for term in terms }
            for m in matchids:

                # alle patterns dieser id
                patterns  = [ term['doc']      for term in terms if term['match_id'] == m]
                callbacks = { term['on_match'] for term in terms if term['match_id'] == m and 'on_match' in term}
                if on_match:
                    callbacks.add(on_match)
                if len(callbacks) == 0:
                    matcher.add( m, patterns )   
                    if verbose:
                        print(m, attr, patterns, 'no callback')                     
                elif len(callbacks) == 1:
                    callback = list(callbacks)[0]
                    matcher.add( m, patterns, on_match=callback ) 
                    if verbose:
                        print(m, attr, patterns, callback )                         
                else:
                    raise ValueError('Only 1 callback per id')
     
                    
        # terms ist Liste von Strings 
        elif rtype_terms == ('list', 'str'):   
            patterns = [  self.nlp.make_doc(term) for term in terms  ]
            matcher.add(match_id, patterns)     
            if verbose:
                print(match_id, attr, patterns)                
            

        else:
            raise ValueError('terms must be list of str or list of dicts')
            
        return matcher
    
    
    
    
    def str2hash( self, text):
        '''Liefert den Hashwert eines Strings'''
        return self.nlp.vocab.strings[text]
    
    def hash2str( self, zahl):
        '''Liefert den String eines Hashwertes'''
        return self.nlp.vocab.strings[zahl]    
    
    
    def tagZ(self, taghash):
        '''Liefert tagZ, ein zusammenfassendes tag.
        Dies ist die Methode blp.tagZ des blp-Objektes, die Hashwerte annimmt.
        Es gibt auch die Funktion bj_nlp.tagZ, die Strings annimmt.
        '''        
        return bj_nlp.tagZ( self.hash2str(taghash) )
    
    
    def tagZZ(self, taghash):
        '''Liefert tagZZ, ein stark zusammenfassendes tag.
        Dies ist die Methode blp.tagZZ des blp-Objektes, die Hashwerte annimmt.
        Es gibt auch die Funktion bj_nlp.tagZZ, die Strings annimmt.
        '''        
        return bj_nlp.tagZZ( self.hash2str(taghash) )    
    
    
    




# Importe
import gensim
import bpyth as bpy

# ==================================================================================================
#  Vereinfachung und Säuberung 
# 

def preprocess_strings( dirty, how=''):
    """ 
    Simplification and cleaning
    * dirty is Series or DataFrame. For a dataframe, all str columns are treated.
    * how determines which purge steps are executed.
      - fillna is always executed anyway
      - strip
      - minus2space replaces hyphens with spaces
      - lower
      - filter_letters deletes everything that is not a letter or number or space
      - umlaut2single replaces umlauts with single letters
      - umlaut2double replaces umlauts with letter combinations          
      - solowhite replaces multiple whitespaces with single space
      - but_anything takes the original if nothing else is left in the end
    * for how there are also presets, these are signalled by an abbreviation

  
    
    
    Vereinfachung und Säuberung
        * dirty ist Series oder DataFrame. Bei einem Dataframe werden alle str-Spalten behandelt.
        * how legt fest, welche Säuberungsschritte ausgeführt werden.
          - fillna wird sowieso immer ausgeführt
          - strip
          - minus2space ersetzt Bindestriche durch Spaces
          - lower
          - filter_letters löscht alles, was nicht Buchstabe oder Zahl oder Space ist
          - umlaut2single ersetzt Umlaute durch Einzelbuchstaben
          - umlaut2double ersetzt Umlaute durch Buchstabenkombinationen          
          - solowhite ersetzt multiple Whitespaces durch einzelnen Space
          - but_anything nimmt dann doch das Original, falls am Ende sonst gar nichts übrig bleibt
        * für how gibt es auch Presets, diese werden durch ein Kürzel signalisiert
    """
    
    # Presets
    if how == '':
        how = 'solowhite strip'
    elif how == 'STD_0': # lower_and_std
        how += ' strip minus2space lower filter_letters solowhite but_anything' 
    elif how == 'GROB_1': # str_grob
        how += ' umlaut2single esszett2ss'       
        
    # Series
    if   type(dirty) == pd.Series:
        
        result      = pd.DataFrame(dirty) 
        result.columns = ['A'] 
        result['A'] = result.A.fillna('') # Kopie, falls für but_anything benötigt
        result['B'] = result.A.copy()
        
        if 'strip'         in how:
            result.B = result.B.str.strip()           
        if 'minus2space'   in how:
            result.B = result.B.str.replace('-', ' ', regex=False)     
        if 'lower'         in how:
            result.B = result.B.str.lower() 
        if 'filter_letters'  in how:
            result.B = result.B.str.replace(r'[^ÄÖÜäüößA-Z a-z0-9]+', '', regex=True)   
        if 'umlaut2single' in how:
            table = str.maketrans({'ä':'a','ö':'o','ü':'u','Ä':'A','Ö':'O','Ü':'U' })
            result.B = result.B.str.translate(table)  
        if 'umlaut2double' in how:
            table = str.maketrans({'ä':'ae','ö':'oe','ü':'ue','Ä':'Ae','Ö':'Oe','Ü':'Ue' })
            result.B = result.B.str.translate(table)              
        if 'esszett2ss'    in how:
            result.B = result.B.str.replace('ß','ss', regex=False) 
        if 'solowhite'     in how:
            result.B = result.B.replace('\s+', ' ', regex=True)                
        if 'strip'         in how:
            result.B = result.B.str.strip()    
        if 'but_anything'      in how:  
            mask = (result.B == '') # nichts übrig geblieben?
            result.loc[mask,'B'] = result[mask].A                 

        return result.B        


    # DataFrame: auf alle Spalten anwenden
    elif type(dirty) == pd.DataFrame:
        
        result = dirty.copy()      
        cols = col_names(result, only='str')
        
        # ausführen
        for col in cols: 
            result[col] = preprocess_strings( result[col], how=how )
        return result
        
    else:
        assert 'wrong datatype'
            
    return result




      

def simplyfy_charset(mystring):
    ''' 
    Vereinfacht den Zeichensatz. 
    Es werden keine Zeichen gelöscht. Aber es können Zeichenkombinationen zu einem Zeichen zusammengefasst werden.
      * öffnende und schließende Anführungszeichen werden zu » und « vereinheitlicht
      * lange und kurze Striche werden zu - vereinheitlicht
      * Pünktchen Pünktchen wird zur Elipse
      * verschiedene Sternchen und Kuller werden zu *
      * diverse Kleinigkeiten
    '''

    ERSETZEN = [('–', '-'),        ('−', '-'),        ('‑', '-'), ('—', '-'), ('‒', '-'), ('‐', '-'), 
                ('›',  '»'),       ('‹',  '«'), 
                ('„',  '»'),       ('“',  '«'),        
                ('‚',  '»'),       ('‘',  '«'),         
                ('“',  '»'),       ('”',  '«'),   
                ('>>', '»'),       ('<<', '«'),   
                ('\u200b', ' '),           
                ('....','…'),      ('...','…'),       ('..','…'),    
                ('œ', 'oe'),       ('Ł','L'),         ('¿', '?'),
                ('❊','*'),         ('✖','*'),         ('•','*'),    ('о','*'),   ('∗','*'),               

               ] 

    for k, v in ERSETZEN:
        mystring = mystring.replace(k, v) 
    return mystring




def remove_bad_chars(mystring, bad_chars=None):
    '''
    Entfernt einige Junk-Zeichen
    '''

    if not bad_chars:
        bad_chars = ['©', '®', '�', '\xad']

    for i in bad_chars :
        mystring = mystring.replace(i, '')
 
    return mystring








def remove_accents(mystring):
    ''' 
    Entfernt Accents. Erhält deutsche Umlaute.
    '''
    
    MASKIEREN = [('ä', '<auml>'),   ('ö', '<ouml>'),   ('ü', '<uuml>'), 
                 ('Ä', '<Auml>'),   ('Ö', '<Öuml>'),   ('Ü', '<Uuml>'), 
                ]     
    
    DEMASKIEREN = [('<auml>','ä'),  ('<ouml>','ö'),    ('<uuml>','ü'), 
                   ('<Auml>','Ä'),  ('<Ouml>','Ö'),    ('<Uuml>','Ü'), 
                   ('`',''),        ('´',''),          ('’', ''),          ('ʹ', ''),  
                ]        
    
    # maskieren
    for k, v in MASKIEREN:
        mystring = mystring.replace(k, v)  
    
    # deaccent
    mystring = gensim.utils.deaccent(mystring)          
    
    # demaskieren    
    for k, v in DEMASKIEREN:
        mystring = mystring.replace(k, v)    
        
    return mystring        
    
    

    
def remove_dupstings(mystring):
    '''
    Entfernt Doppelungen wie z.B. in "Vereinigtes Königreich Vereinigtes Königreich"
    '''
    return ' '.join(bpy.remove_dups(mystring.split(' ')))    




def preprocess(text, preserve_paragraphs=False, config=0):
    '''
    Wendet eine Liste von Funktion nacheinander auf einen String an
    '''
    
    if preserve_paragraphs:
        return '\n'.join(  [preprocess(absatz, preserve_paragraphs=False, config=config) for absatz in text.split('\n') if len(absatz) > 1]   )    

    CONFIG_0 = [simplyfy_charset,
                remove_bad_chars,
                remove_accents, 
               ]
  
    if config == 0:
        config = CONFIG_0   
    elif config == 1:
        config = CONFIG_1       
    
    result = ' '.join(  gensim.parsing.preprocessing.preprocess_string(  str(text), config)  )    
    return result



  
    

                
                

      



import pandas as pd

from bj_nlp            import blpClass
from spacy.tokens      import Doc, Token, Span






class WRede:
    '''
    Pipeline-Komponente
    wrede
    ========
    Liefert folgende Custom Attibutes:
    - doc._.wrede          (eine Liste von Spans)
    - token._.is_wrede     (Integer, Anzahl der zusammenhängenden Token die wörtliche Rede sind, 
                            0 wenn außerhalb wörtlicher Rede)

    Registrierung siehe blpClass.
    
    # Aktivierung:
    if not 'wrede' in nlp.pipe_names:
        nlp.add_pipe('wrede',
                     before='parser',            # vor wortschatz
                     config={'debug': False} 
                    )   
                
    # direkter Zugriff auf die Pipeline-Komponente:
    wrede = blp.nlp.get_pipe('wrede') 
    
    .
    '''    
    
    # ===========================================================================================0 
    # init
    #    
    
    def __init__(self, blp: blpClass, debug: bool):        
        
        self.blp   = blp                # das blp-Objekt
        self.name  = 'wrede'            # name, für Error-Messages     
        self.debug = debug        
        
        # Create the matcher    
        rules = []
        rules += [{  'match_id': 'WREDE_L',   'text': '»',  }]
        rules += [{  'match_id': 'WREDE_R',   'text': '«',  }]  
        rules += [{  'match_id': 'WREDE',     'text': '"',  }]             
        self.matcher = blp.phrasematcher(rules, verbose=False )
        
        # Register custom extension on the Doc
        if not Doc.has_extension('wrede'):
            Doc.set_extension('wrede', default=[])
            
        if not Token.has_extension('is_wrede'):
            Token.set_extension('is_wrede', default=0)            
            
        if debug:            
            if not Doc.has_extension('debug'):
                Doc.set_extension('debug', default=[])            
                        
        if self.blp.verbose:
            print( self.name + ' initialisiert')            

        
    
   
    
    # ===========================================================================================0
    # Hauptroutine
    #    
    
    def __call__(self, doc:Doc) -> Doc:
        
        # Finden
        matches = self.matcher(doc)
        
        if len(matches) == 0:
            return doc
        
        # in DataFrame wandeln
        df = self.blp.df_matches(matches, as_string=True)
        
        # Passende Anführungszeichen finden
        df['Anfang']          = df.start.shift(1).astype('Int32')
        df['Anfang_match_id'] = df.match_id.shift(1)      
        
        # Jeden zweiten Datensatz löschen

        
        # Unpassende löschen
        maskA = (df.Anfang_match_id == 'WREDE_R')  &  (df.match_id == 'WREDE_L')
        maskB = (df.Anfang_match_id == 'WREDE')    &  (df.match_id != 'WREDE')    
        maskC = (df.match_id == 'WREDE')           &  (df.Anfang_match_id != 'WREDE')     
        maskD = (df.Anfang_match_id == 'WREDE_L')  &  (df.match_id == 'WREDE_L')
        maskE = (df.Anfang_match_id == 'WREDE_R')  &  (df.match_id == 'WREDE_R')        
        maskF =  df.Anfang_match_id.isnull()
        mask = maskA | maskB  |  maskC  |  maskD  |  maskE  |  maskF
        df = df[~mask]        
        
        #df = df.iloc[1::2, :]
        
        if self.debug:
            doc._.debug = df
        
        # is_wrede
        for index, row in df.iterrows():
            for i in range(row.Anfang+1, row.end-1):
                doc[i]._.is_wrede = row.end-row.Anfang-2       
        
      
        # Funktion Span heraussuchen
        def collect_wrede(zeile):
            zeile['wrede'] = doc[zeile.Anfang+1 : zeile.end-1]
            return zeile    
        
        # apply Span heraussuchen
        df['wrede'] = ''
        df = df.apply(collect_wrede, axis=1)        
           
        # Ergebnisse speichern    
                
        doc._.wrede = list(df.wrede)
        
        return doc
    
    
    
    # ===========================================================================================0 
    # testtext
    #
    
    TESTTEXT_0 = 'Kap›utt ›Hallo!‹ ›Wie geht es dir?‹ "Mir?" ››Mir geht es gut.‹ Kap"utt'    
    
    TESTTEXT_1 = '''Leni fragt: »Kannst du dir das alles merken?«
Sascha sagt: »Das Spiel gefällt mir gut.«
Der Vater befiehlt: »Mach jetzt endlich deine Hausaufgaben!«
»Kannst du dir das alles merken?«, fragt Leni.
»Das Spiel gefällt mir gut«, sagt Sascha.
»Mach jetzt endlich deine Hausaufgaben!«, befiehlt der Vater.
»Kannst du«, fragt Leni, »dir das alles merken?«
»Das Spiel«, sagt Sascha, »gefällt mir gut.«
»Mach jetzt«, befiehlt der Vater, »endlich deine Hausaufgaben!«
»M'''
    

    
  
    
    TESTTEXTE = {0:TESTTEXT_0, 1:TESTTEXT_1}
    

    
    
    # ===========================================================================================0 
    # tests durchführen
    #
    def tests(self, only=None, verbose=True):  
        
        # Leeres Dokument
        doc = self.blp.nlp('')
        assert len( list(doc.sents)  ) == 0        
        
        # Anzahl der Sätze in TESTTEXT_0 prüfen            
        if only is None or only == 0:
            
            text_0 = self.TESTTEXTE[0]
            text_1 = self.blp.preprocess(text_0)
            doc = self.blp.nlp(text_1)
            assert len(list(doc.sents)) == 5
            assert len(doc._.wrede) == 4
            if verbose:
                print('wrede Test 0 OK')           
        
        
        
        # Anzahl der Sätze in TESTTEXT_1 prüfen
        if only is None or only == 1:    

            text_0 = self.TESTTEXTE[1]
            text_1 = self.blp.preprocess(text_0)
            doc = self.blp.nlp(text_1)
            assert len(list(doc.sents)) == 10
            assert len(doc._.wrede) == 12
            if verbose:
                print('wrede Test 1 OK')    
             
            
            
             
    
    
    
    
        
        




#####################################################################################################
# Schnittstellen zu Pandas
#####################################################################################################     
   
import spacy
from spacy.tokens         import Doc, Token, Span    
    
try:
    from dframcy import DframCy
except:
    pass
    
import pandas     as pd
import pandasklar as pak     
    
#from bj_nlp.translate_tags  import *



class PandaziseMixin:

    
    def df_token(self, doc, columns = ['BASIC']):
        '''
        Liefert einen DataFrame mit den gewünschten Feldern.
        Input ist eine Sequenz von Token, z.B. ein doc oder ein subtree.
        An Pseudofeldern stehen ['BASIC','HELP','LING'] zur Verfügung, diese werden jeweils durch ein Defaultset von Feldern ersetzt.
        Wenn möglich, wird statt eines Hashwertes ein String ausgegeben.
        Will man explizit den Hashwert statt des Strings, so verwendet man den Funktionsnamen mit angehängtem '_hash'. 
        '''
        
        cols = []
        if 'BASIC' in columns:
            cols += ['text', 'lemma', 'tag', 'pos']   
                
        if 'ENT' in columns:
            cols += ['ent_type', 'ent_iob_int',]                 
                
        if 'DEP' in columns:
            cols += ['text', 'pos', 'tag', 'dep', 'dep_help', 'head', 'head_pos', 'ancestors', 'children',] 
            
        if 'HELP' in columns:
            cols += ['tag_help', 'pos_help', 'dep_help']       
            
        if 'CHAR' in columns:
            cols += ['text', 'pos', 'is_punct', 'is_left_punct', 'is_right_punct', 'is_sent_start'] 
            
        cols += columns            
        cols = [  c for c in cols if c not in ['BASIC','ENT','DEP','HELP','CHAR']  ]    
        cols = list(dict.fromkeys(cols))   # remove dups keep order    
        #print('cols=',cols)
        
        # Daten zusammenstellen
        result_raw = []
        for token in doc:
            zeile = list( self.token_xxx(doc, token, symbol )  for symbol in cols )
            result_raw.append(zeile)

        # DataFrame liefern
        result = pd.DataFrame(result_raw, columns=cols) 
        
        # Reihenfolge
        if 'pos_help' in result.columns:
            result = pak.move_cols(result, 'pos_help', 'pos')            
        if 'tag_help' in result.columns:
            result = pak.move_cols(result, 'tag_help', 'tag')            
        if 'dep_help' in result.columns:
            result = pak.move_cols(result, 'dep_help', 'dep')            
            
        return result    
    
    
    
       
    

    # Hilfsroutine
    def token_xxx( self, doc, token, func_name ):
        
        # Liefert zu einem Token eine Liste von ancestor-Lemmas, ggf. gefiltert
        def ancestor_lemmas(token, pos=''):
            if pos == '':
                result = [token.lemma_ for token in token.ancestors]  
            else:
                result = [token.lemma_ for token in token.ancestors if token.pos_ == pos]  
            return result    
        
        # Liefert zu einem Token eine Liste von ancestors, ggf. gefiltert
        def ancestors(token, pos=''):
            if pos == '':
                result = [token.text for token in token.ancestors]  
            else:
                result = [token.text for token in token.ancestors if token.pos_ == pos]  
            return result            
        
        # Liefert zu einem Token eine Liste von children-Lemmas, ggf. gefiltert
        def children_lemmas(token, pos=''):
            if pos == '':
                result = [token.lemma_ for token in token.children]  
            else:
                result = [token.lemma_ for token in token.children if token.pos_ == pos]  
            return result        
        
        # Liefert zu einem Token eine Liste von children-Lemmas, ggf. gefiltert
        def children(token, pos=''):
            if pos == '':
                result = [token.text for token in token.children]  
            else:
                result = [token.text for token in token.children if token.pos_ == pos]  
            return result          
        
        def ent_iob_int(token):
            t = {'B':1, 'I':2, 'O':0}
            return t[token.ent_iob_]
        
        # text
        if func_name == 'text':
            return token.text  
        
        # ent 
        elif func_name == 'ent_iob_int':              # 0=outside, 1=start, 2=inside          
            return ent_iob_int(token)           
        
        # Tree
        elif func_name == 'head_pos':                # The part-of-speech tag of the token head
            return token.head.pos_               
        elif func_name == 'children':                # The immediate syntactic dependents of the token
            return children(token) 
        elif func_name == 'children_lemmas':               
            return children_lemmas(token)         
        elif func_name == 'ancestors':                
            return ancestors(token)
        elif func_name == 'ancestor_lemmas':                
            return ancestor_lemmas(token)        
        
        elif func_name == 'verben':
            return ancestors(token,'VERB')    
        elif func_name == 'adjektive':
            return ancestors(token,'ADJA')    
        
        
       # elif func_name == 'sent':                  # sent ist in bj_tools realisiert
       #     return token.sent.text.strip()    
       # elif func_name == 'sent_pre':
       #     satz_span = token.sent.text.strip()   
       #     return doc[satz_span.start-1].sent           
        
        # HELP
        elif func_name == 'dep_help':
            return spacy.explain(token.dep_) 
        elif func_name == 'pos_help':
            return spacy.explain(token.pos_)    
        elif func_name == 'tag_help':
            return tag_help(token.tag_)   
        
        # morph
        elif func_name == 'morph_dict':
            return token.morph.to_dict()        
        elif func_name.startswith('morph_list_'):   # z.B. morph_Person liefert Person des morph-Dictionary als Liste
            return token.morph.get(func_name[11:])
        elif func_name.startswith('morph_'):        # z.B. morph_Person liefert Person des morph-Dictionary als Skalar. 
            m = token.morph.get(func_name[6:])      # Falls es mehrere Werte sind, dann nur den ersten Wert. 
            if len(m) > 0:
                return m[0]
            else:
                return None
    
        # standard attributes: Versucht stets str zurückzugeben, wenn möglich
        else:
            
            # Eingebautes Attribut, str
            if hasattr(token, func_name+'_'):
                try:
                    return getattr(token, func_name+'_')
                except:
                    return 'Error: ' + func_name+'_'                 
            
            # Eingebautes Attribut, hash            
            if func_name[-5:] == '_hash' and hasattr(token, func_name[:-5] ):     
                try:
                    return getattr(token, func_name[:-5])                     
                except:
                    return 'Error: ' + func_name[:-5]                 
            
            # Eingebautes Attribut
            if hasattr(token, func_name):
                try:
                    return getattr(token, func_name).text            
                except:
                    try:
                        return getattr(token, func_name)  
                    except:
                        return 'Error: ' + func_name
                    
            
            # Selbst definiertes Attribut, str
            if hasattr(token._, func_name+'_'):
                try:
                    return getattr(token._, func_name+'_')
                except:
                    return 'Error: ' + '_.'+func_name+'_'                     
            
            # Selbst definiertes Attribut, hash            
            if func_name[-5:] == '_hash' and hasattr(token._, func_name[:-5] ):     
                try:
                    return getattr(token._, func_name[:-5])                     
                except:
                    return 'Error: ' + '_.'+func_name[:-5]                  
            
            # Selbst definiertes Attribut            
            if hasattr(token._, func_name):
                try:
                    return getattr(token._, func_name)                     
                except:
                    return 'Error: ' + '_.'+func_name      
                

            
            
            return 'ERRORi: ' + func_name
                
           

        
        
        




        
        
        
    # Liefert einen DataFrame mit den gewünschten Feldern
    #
    def df_span(self, spans, columns=['BASIC']):
        '''
        Liefert einen DataFrame mit den gewünschten Feldern.
        Input ist eine Sequenz von Token, z.B. doc.ents
        An Pseudofeldern steht nur ['BASIC'] zur Verfügung, diese werden jeweils durch ein Defaultset von Feldern ersetzt.
        '''        
        cols = []
        if 'BASIC' in columns:
            cols += ['start', 'end', 'text', 'label', ]   
            if Token.has_extension('meta'): 
                cols += ['meta']
            if Token.has_extension('info'):
                cols += ['info']
            
        cols += columns            
        cols = [  c for c in cols if c not in ['BASIC','LING','HELP']  ]    
        cols = list(dict.fromkeys(cols))   # remove dups keep order    
        #print('cols=',cols)        

        # Daten zusammenstellen
        result_raw = []
        for span in spans:
            zeile = list( self.span_xxx(spans, span, symbol )  for symbol in cols )
            result_raw.append(zeile)

        # DataFrame liefern
        result = pd.DataFrame(result_raw, columns=cols) 
        return result         
        
        

    # Hilfsroutine
    def span_xxx( self, spans, span, func_name ):
        
        # text
        if func_name == 'text':
            return span.text   
        
        # Tree
        elif func_name == 'root_dep':         # Dependency relation connecting the root to its head
            return span.root.dep_        
        elif func_name == 'root_head':        # The text of the root token’s head
            return span.root.head.text           
        
        # extensions
        elif func_name == 'meta':
            print('TODO: in General-Zweig unten einbauen')
            return spans[span.start]._.meta   
        elif func_name == 'info':
            print('TODO: in General-Zweig unten einbauen')
            return spans[span.start]._.info    
            
        # standard attributes: Versucht stets str zurückzugeben, wenn möglich
        else:
            if hasattr(span, func_name+'_'):
                return getattr(span, func_name+'_')
            if hasattr(span, func_name):
                try:
                    return getattr(span, func_name).text            
                except:
                    try:
                        return getattr(span, func_name)  
                    except:
                        return 'Error: ' + func_name
            if hasattr(span._, func_name):
                try:
                    return getattr(span._, func_name) 
                except:
                    pass
            
            return 'ERROR: ' + func_name                    
                    
          

        
        
        
    # Liefert einen DataFrame mit den gewünschten Feldern
    #
    def df_lex(self, doc, columns=['BASIC']):
        '''
        Liefert einen DataFrame mit den gewünschten Feldern.
        Input ist eine Sequenz von Token, z.B. doc.ents
        An Pseudofeldern steht nur ['BASIC'] zur Verfügung, diese werden jeweils durch ein Defaultset von Feldern ersetzt.
        '''        
        cols = []
        if 'BASIC' in columns:
            cols += ['text', 'is_alpha', 'is_digit','is_punct', 'is_title', 'shape']     
        if 'CHAR' in columns:
            cols += ['text', 'is_punct', 'is_left_punct', 'is_right_punct','is_bracket','is_quote']
            
        cols += columns            
        cols = [  c for c in cols if c not in ['BASIC','CHAR','HELP']  ]    
        cols = list(dict.fromkeys(cols))   # remove dups keep order    
        #print('cols=',cols)        

        # Daten zusammenstellen
        result_raw = []
        for token in doc:
            lex = doc.vocab[token.text]
            zeile = list( self.lex_xxx(lex, symbol )  for symbol in cols )
            result_raw.append(zeile)

        # DataFrame liefern
        result = pd.DataFrame(result_raw, columns=cols) 
        return result         
        
        

    # Hilfsroutine
    def lex_xxx( self, lex, func_name ):
        
        # text
        if func_name == 'text':
            return lex.text   
        
        # standard attributes: Versucht stets str zurückzugeben, wenn möglich
        else:
            if hasattr(lex, func_name+'_'):
                return getattr(lex, func_name+'_')
            if hasattr(lex, func_name):
                try:
                    return getattr(lex, func_name).text            
                except:
                    try:
                        return getattr(lex, func_name)  
                    except:
                        return 'Error: ' + func_name
            if hasattr(lex._, func_name):
                try:
                    return getattr(lex._, func_name) 
                except:
                    pass
            
            return 'ERROR: ' + func_name                

        
           
    def analyze_pipes(self, pretty=False):
        '''
        Wie spacys analyze_pipes, liefert aber DataFrame
        '''
        analyse = self.nlp.analyze_pipes(pretty=pretty)
        result = pak.dataframe(  analyse['summary']   ).transpose()        
        result = pak.reset_index( result,  keep_as='component' )     
        if pretty:
            print()
        for key, value in analyse['problems'].items():
            if len(value) > 0:
                print('Problem in {0}: {1}'.format(key, value))   
        
        # Info printen
        print(  'model_name:', self.model_name       )
        print(   'vocab_len:', len(self.nlp.vocab)   )
        
        
        
        if len(self.nlp.disabled) > 0:
            print('pipe component disabled:',self.nlp.disabled)
                
        return result
    
    
    
    def df_matches( self, matches, as_string=True ):
        '''
        Converts the result of a Matcher to DataFrame
        '''
        result = pd.DataFrame(matches)
        result.columns = ['match_id','start','end']
        if as_string:
            result['match_id'] = result.match_id.apply( self.hash2str )
        return result        
        
        

    
    
    





# splittet ein lexem in einen Kernbereich und eine Differenz
def diff_grammatik(lemma, lex):
    
    
    # splittet ein lexem in einen Kernbereich und eine Differenz
    def diff_grammatik_intern(lemma, lex):
        for i, (char1, char2) in enumerate(zip_longest(lemma, lex)):
            if char1 != char2:
                if i == 0:
                    return longest_substr([lemma, lex]),'' # 
                else:
                    return lex[:i],lex[i:] # normale Ausgabe
        return lemma,'' # Strings sind identisch    
    
    
    # main
    if lemma == lex:
        return lex, ''
    basis1, zusatz1 = diff_grammatik_intern( lemma,       lex )
    #if zusatz1=='☆': # ohne Umlautproblem
     #   return basis1, zusatz1
    
    lemma_ascii = lemma.replace('ä','a').replace('ö','o').replace('ü','u')
    lex_ascii   =   lex.replace('ä','a').replace('ö','o').replace('ü','u')   
    if lemma_ascii == lex_ascii:
        return lex, ''
    
    basis2, zusatz2 = diff_grammatik_intern( lemma_ascii, lex_ascii )  
    
    if len(basis2) <= len(basis1)   and   zusatz1 != '': #d ohne Umlautproblem
        return basis1, zusatz1

    if zusatz2 == ''  or zusatz1 == '':  #or  zusatz2 == '☆':
        return lex, ''
    else: 
        return lex[:len(basis2)], lex[-len(zusatz2):]






# TODO: Punkt und Großgeschriebenes Adjektiv o.ä. weist auf Satzende hin.

#import pandas as pd

from bj_nlp            import blpClass
from spacy.tokens      import Doc, Token  #, Span
from spacy.matcher     import Matcher
#from bpyth             import load_pickle


# Funktionen tagZ, tagZZ
from bj_nlp.translate_tags  import *





class Wortschatz:
    '''
    Pipeline-Komponente
    wortschatz
    ==========
    Schlägt im Wiktionary nach und stellt einige Custom Attributes zur Verfügung:
    - Token._.tagZ und tagZZ
    - Token._.sent und sent_pre  
    - Token._.lemmaW_ und lemmaW
    Außerdem werden die Satzgrenzen neu geschrieben:
    - Token.is_sent_start


    Registrierung siehe blpClass.
    
    # Aktivierung:
    if not 'wortschatz' in nlp.pipe_names:
        nlp.add_pipe('wortschatz', 
                     before='parser',             # in jedem Fall vor dem Parser
                     config={'debug': True} 
                     ) 
        
    # direkter Zugriff auf die Pipeline-Komponente:
    wortschatz = blp.nlp.get_pipe('wortschatz') 
    
    .
    '''    
    
    # ===========================================================================================0 
    # init
    #    
    
    def __init__(self, blp: blpClass, debug: bool):
        
        
        self.blp   = blp                    # das blp-Objekt
        self.name  = 'wortschatz'           # name, für Error-Messages 
        self.debug = debug
        
        self.sent_end   = ['.',':','!','?','«']   # Zeichen, die Satzende bedeuten         
        self.sent_start = ['»']                   # Zeichen, die baldigen Satzanfang bedeuten 
                    
        if debug:
            if not Doc.has_extension('debug'):
                Doc.set_extension(   'debug', default='' )     

            if not Token.has_extension('debug'):
                Token.set_extension(   'debug', default='')   
            
            
        if not Token.has_extension('tagZ'):
            Token.set_extension(   'tagZ',  getter=self.get_tagZ)
            
        if not Token.has_extension('tagZZ'):
            Token.set_extension(   'tagZZ', getter=self.get_tagZZ)          
            
            
        if not Token.has_extension('sent'):
            Token.set_extension(   'sent', getter=self.get_sent)     
            
        if not Token.has_extension('sent_pre'):
            Token.set_extension(   'sent_pre', getter=self.get_sent_pre)               
            
                        
        if not Token.has_extension('lemmaW_'):
            Token.set_extension(   'lemmaW_',   default = '' )   
            
        if not Token.has_extension('lemmaW'):
            Token.set_extension(   'lemmaW',  getter=self.get_lemmaW )               
                         
        if not Token.has_extension('tagW_'):
            Token.set_extension(   'tagW_',   default = '' )   
            
        if not Token.has_extension('tagW'):
            Token.set_extension(   'tagW',  getter=self.get_tagW )     
            
        if not Token.has_extension('member'):
            Token.set_extension(   'member',  default = '' )             
            

        # load lookups
        try:
            import importlib.resources as pkg_resources
        except ImportError:
            # Try backported to PY<37 `importlib_resources`.
            import importlib_resources as pkg_resources
        from . import data     

        import pickle
        text = pkg_resources.read_binary( data, 'lookup1.pickle' )
        self.lookup1 = pickle.loads(text)      
        text = pkg_resources.read_binary( data, 'lookup2.pickle' )
        self.lookup2 = pickle.loads(text) 
        text = pkg_resources.read_binary( data, 'lookup3.pickle' )
        self.lookup3 = pickle.loads(text) 
        text = pkg_resources.read_binary( data, 'lookup4.pickle' )
        self.lookup4 = pickle.loads(text)      
            
            
        # === Matcher zur Erkennung von Abkürzungen
        self.matcher = Matcher(blp.nlp.vocab)
        patterns  = []
        
        # Ein oder zwei Zeichen, dann Punkt, dann wieder ein oder zwei Zeichen 
        token1 = {"TEXT": {"REGEX": "^.{1,2}\..{1,2}$"}}
        token2 = {"ORTH": "."}
        patterns +=  [[  token1, token2 ]] 
        
        # Ein oder zwei Konsonanten
        wird_nicht_mit_punkt_abgekürzt = ['au','ei','hm','qm','ui','mg','km','cm','mm','dm','ft','nm',
                                          'pt','m²','m³','kg','lb','db','BH','CD','CC','CT','DB','DC','DM','FH','KW','VW','XY']
        token1 = {"TEXT": {"REGEX": "^[b-df-hj-np-tv-zßB-DF-HJ-NP-TV-Z]{1,2}$"},  
                  "ORTH": {"NOT_IN": wird_nicht_mit_punkt_abgekürzt}
                 }           
        token2 = {"ORTH": "."}
        patterns +=  [[  token1, token2 ]]          
        
        #add patterns
        self.matcher.add('abbrev', patterns)  
        
        
        # === Matcher zur Wiedererkennung von Abkürzungen
        self.matcher2 = Matcher(blp.nlp.vocab)
        patterns  = []
        
        # Wort mit Punkt
        token1 = {"TEXT": {"REGEX": "\w+\."}}
        patterns +=  [[  token1 ]]      
        
        #add patterns
        self.matcher2.add('no_sentstart', patterns)          
        
        
        
        if self.blp.verbose:
            print( self.name + ' initialisiert')           
        

        
    
    
    # ===========================================================================================0
    # getter
    #      
    
    def get_tagZ(self,token):
        return self.blp.tagZ(token.tag)       
    
    
    
    def get_tagZZ(self,token):
        return self.blp.tagZZ(token.tag)       
    
    
    
    def get_sent(self,token):
        return token.sent.text.strip()    
    
    
    
    def get_sent_pre(self,token):

        i = token.sent.start-1
        if i > 0:
            return token.doc[i]._.sent 
        else:
            return ''    
    
    
    
    def get_lemmaW(self,token):
        try:
            return self.blp.str2hash(  token._.lemmaW_ )   
        except KeyError:
            if self.debug:
                token._.debug += '↯'
            return -7 

        
        
    def get_tagW(self,token):
        try:
            return self.blp.str2hash(  token._.tagW_ )   
        except KeyError:
            if self.debug:
                token._.debug += '↯'
            return -7         
    
    
    
    # ===========================================================================================0
    # Hauptroutine: 
    # - belegt lemmaW_ und tagW_
    # - joined Token (retokenize)
    # - schreibt is_sent_start
    #    
    def __call__(self, doc:Doc) -> Doc:
    
        
        if len(doc) <= 1:
            return doc  
        

        
        # --------------------------------------------------------
        # Turn 0: Abkürzungen per Matcher suchen
        #               
            
        matches = self.matcher(doc)
        spans = []  # Collect the matched spans here
        for match_id, start, end in matches:
            spans.append(doc[start:end])
            if self.debug:
                doc[end]._.debug += ' ↰'            
            doc[end].is_sent_start = False

        with doc.retokenize() as retokenizer:
            for span in spans:
                retokenizer.merge(span)          
        
     
        
        # --------------------------------------------------------
        # Turn 1: Retokenize laut Wörterbuch
        #          
        
        with doc.retokenize() as retokenizer:
            
            verschmolzen    = ''  
            verschmolzen_id = -1
            
            # Doppelwörter suchen 
            for i,token in enumerate(doc):
                    
                if i >= len(doc)-1:
                    continue
                    
                if doc[i].tag_ == '_SP':
                    continue
                    
                if doc[i+1].tag_ == '_SP':
                    continue                                    
                    
                span      = doc[i:i+2]  
                span_hash = self.blp.str2hash( span.text.lower() )
                eintrag   = self.lookup1.get( span_hash, None )      
                
                # Zweiter versuch
                if not eintrag: 
                    eintrag   = self.lookup4.get( span_hash, None ) 
                    
                # Immer noch nichts gefunden >> weiter                      
                if not eintrag: 
                    continue                    
                    
                # gefunden >> zu einem Token verschmelzen                                       
                # für Errors merken    
                verschmolzen_id_pre = verschmolzen_id
                verschmolzen_id     = i
                verschmolzen_pre    = verschmolzen    
                verschmolzen        = span                    

                if self.debug:
                    doc[i]._.debug     += ' ⨝'              
                doc[i]._.lemmaW_       = eintrag[0]     
                doc[i]._.tagW_         = eintrag[1]   

                # verschmelzen
                try: 
                    retokenizer.merge(span) 
                except ValueError:         
                    error = {'error'    : 'Überlappende Verschmelzungen',
                            'component' : self.name, 
                             'nam0'     : 'verschmolzen_pre',
                             'val0'     : verschmolzen_pre,  
                             'nam1'     : 'verschmolzen',
                             'val1'     : verschmolzen,      
                             #'nam2'     : 'doc_id',
                             #'val2'     : self.spacy_batch.doc_id,          
                             'nam3'     : 'pos',
                             'val3'     : i                                
                            }
                    if self.debug:
                        doc[i]._.debug += str(error)                          
        
        # retokenizer end
        
        
        
        # --------------------------------------------------------
        # Turn 2: Abkürzungen wiederfinden
        #               
            
        matches = self.matcher2(doc)
        for match_id, start, end in matches:
            try:
                doc[end]._.debug += ' ⧉'
                doc[end].is_sent_start = False
            except:
                pass
            #

       
          
        
        
        
        # --------------------------------------------------------
        # Turn 3: Einzelwörter suchen und Sentencizer
        #    
        
        # DIESE ZEILEN ERZEUGEN EXTREME CPU_LAST
        #for i,token in enumerate(doc): 
            #doc.c[i].sent_start = -1
            #doc[i].is_sent_start   = False 
        
        for i,token in enumerate(doc):      
            
            # Custom Sentencizer: Zeichen für Satzende berücksichtigen
            if i <= len(doc)-2:

                if token.text in self.sent_start:
                    doc[i].is_sent_start = True   
                    try:
                        doc[i+1].is_sent_start   = False                        
                        doc[i-1].is_sent_start   = False  
                    except:
                        pass
                    
                elif token.text in self.sent_end:
                    doc[i+1].is_sent_start = True 
                    if self.debug:
                        doc[i+1]._.debug += ' ⒠' 
                    doc[i].is_sent_start   = False                    
                    
                else:
                    # DIESE ZEILE ERZEUGT EXTREME CPU_LAST
                    #doc[i+1].is_sent_start = False  # Tell the default sentencizer to ignore this token   
                    #doc[i+1].is_sent_start = None 
                    pass

      
            
            # lemmaW_ schon belegt
            if token._.lemmaW_:
                continue
            
            # in lookup1 suchen, d.h. die eindeutigen Fälle
            eintrag = self.lookup1.get( token.lower, None )  
            if eintrag:
                if self.debug:
                    token._.debug += ' ⑴'
                token._.lemmaW_    = eintrag[0]
                token._.tagW_      = eintrag[1]               
                continue
                
            # in lookup2 suchen, d.h. nach lex_lower und tagZ           
            suchetuple = (token.lower, token._.tagZ)
            eintrag = self.lookup2.get( suchetuple, None )  
            if eintrag:
                if self.debug:
                    token._.debug += ' ⑵'
                token._.lemmaW_    = eintrag[0]
                token._.tagW_      = eintrag[1]                     
                continue  
                    
            # in lookup3 suchen, d.h. nach lex_lower und tag           
            suchetuple = (token.lower, token.tag_)
            eintrag = self.lookup3.get( suchetuple, None )  
            if eintrag:
                if self.debug:
                    token._.debug += ' ⑶'
                token._.lemmaW_    = eintrag[0]
                token._.tagW_      = token.tag_                    
                continue     
                
            # in lookup4 suchen, d.h. bei großen Unterschieden zwischen Spacys lemma und Wiktionarys lemma
            eintrag = self.lookup4.get( token.lower, None )  
            if eintrag:
                if self.debug:
                    token._.debug += ' ⑷'
                token._.lemmaW_    = eintrag[0]
                token._.tagW_      = eintrag[1]               
                continue                
                    

                    
            # end for token
            
            
          
                
        return doc
                
                
 
    
    
    
    
    
    
    
    
    
    
    # ===========================================================================================0 
    # Tests
    #
    
    
    TESTTEXT_0 = '''Mrs. Summer kauft eine CD.
Fast tägl. kauft Mrs. Summer neue CDs!
Tägl. kauft Mrs. Summer neue CDs.
Hallo Hn.g. so ist das.'''
    
    TESTTEXT_1 = 'Lena z. läuft z. B. sie rennt z.B. sie d.h. sie d. h. rennt gbRteWq.'
    
    TESTTEXT_2 = '''Wie geht es dir z.z. ja. 
Sein Satz. 
Hallo gs. Bär es sind 5 mm.
U.a. mag ich Butter und Tee, am liebsten tägl. und heute auch.
Heute geschl. außer 500 ca. n.Chr. etc. bzw. Ende.
Auch u.a. deswegen, weil was denkt der Herr z.B. über dies oder u.a. diese Abkürzung, fragte Prof. Dr. Blöd.
Was denkst du z. B. oder u. a. diese Abkürzung?
D.h. wenn du willst.
D. h. wenn du willst.
Ja, d.h. nein, doch nicht.
Ja, d. h. nein, doch nicht.
Ja, tgl. bedeutet tägl., wissen Mr. Spock und Mrs. Spock.
Kennen Dr. Dumm und denn z.B. diese Abkürzung?
Z.B. diese hier.
Hat Prof. Selering die Tür geschlossen?
Tägl. Dusche erforderlich.
Bzw. eine Handwäsche.
U.a. mit Haare usw. waschen.
Ca. drei Stunden!
Nr. 5 lebt.
Auch Nr.3 erfreut sich bester Gesundheit.
Ja, Mrs. Miller sieht das genauso.'''
    
    zweierworte =  ['ab','am','an','aß','da','du','eh','er','es','he','im','in','ja','je','la','mg',]
    zweierworte += ['ne','ob','oh','on','so','tu','um','wo','zu','äh','au','bi','ei','ex','ey','go','hi',]
    zweierworte += ['hm','ho','hä','hü','is','me','mi','na','nu','nö','or','qm','re','si','sä',]
    zweierworte += ['to','uh','ui','up','uz','wa','we','öd','öl','ös','üb','μm']
    zweierworte = [z+'.' for z in zweierworte]
    TESTTEXT_3 = ' '.join(zweierworte) #+ ' ' + ''.join(zweierworte)       
        
    
    TESTTEXT_4 = '''Leni fragt: »Kannst du dir das alles merken?«
Sascha sagt: »Das Spiel gefällt mir gut.«
Der Vater befiehlt: »Mach jetzt endlich deine Hausaufgaben! Wie oft soll ich das noch sagen?«
»Kannst du dir das alles merken? Ich nicht«, sagt Leni.
»Das Spiel gefällt mir gut! Dir auch?«, fragt Sascha.
»Mach jetzt endlich deine Hausaufgaben!«, befiehlt der Vater.
»Kannst du«, fragt Leni, »dir das alles merken?«
»Das Spiel«, sagt Sascha, »gefällt mir gut.«
»Mach jetzt«, befiehlt der Vater, »endlich deine Hausaufgaben!«
»M'''    
    
    
    TESTTEXT_5 ='''U.a. mag ich Butter und Tee tägl. geschl. n.Chr. ca. etc. bzw. Ende.
U.a. was denkt der Herr z.B. über dies oder u.a. diese Abkürzung?
, fragte Prof. Dr. Blöd.
Was denkst du z. B. oder u. a. diese Abkürzung?
D.h. wenn du willst.
D. h. wenn du willst.
Ja, d.h. nein, doch nicht.
Ja, d. h. nein, doch nicht.
Ja, tgl. bedeutet tägl., wissen Mr. Spock und Mrs. Spock.
Kennen Dr. Dumm und denn z.B. diese Abkürzung?
Z.B. diese hier.
Hat Prof. Selering die Tür geschlossen?
Tägl. Dusche erforderlich.
Bzw. eine Handwäsche.
U.a. mit Haare usw. waschen.
Ca. drei Stunden!
Nr. 5 lebt.
Auch Nr.3 erfreut sich bester Gesundheit.
Ja, Mrs. Miller sieht das genauso.
Der Schlusssatz steht am Ende, inkl. Punkt.
'''            
    
    
    
    TESTTEXTE = { 0:TESTTEXT_0, 1:TESTTEXT_1, 2:TESTTEXT_2, 3:TESTTEXT_3, 4:TESTTEXT_4, 5:TESTTEXT_5, }
    

    
    
    # ===========================================================================================0 
    # tests durchführen
    #
    def tests(self, only=None, verbose=True):           
        
        # Test 0: doc erstellen
        text = self.TESTTEXTE[0]
        sent_anz = text.count('\n')+1
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz
        
        # sent und sent_pre
        assert doc[2]._.sent   == 'Mrs. Summer kauft eine CD.'
        assert doc[7]._.sent   == 'Fast tägl. kauft Mrs. Summer neue CDs!'          
        assert doc[17]._.sent  == 'Tägl. kauft Mrs. Summer neue CDs.'   
        assert doc[23]._.sent  == 'Hallo Hn.g. so ist das.'
        
        # tagZ und tagZZ
        assert doc[2]._.tagZ == 'VER'
        assert doc[2]._.tagZZ == 'V'    
        
        # lookup1    
        test = self.lookup1[  7608359307979301264  ]    
        assert test[0] == 'heimtrauen'        
        
        if verbose:
            print('wortschatz Tests 0 OK')              
        
        
        # Test1: doc erstellen
        text = self.TESTTEXTE[1]
        sent_anz = text.count('\n')+1        
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz        
        
        # lemmaW
        assert doc[0]._.lemmaW_ == 'Lena'
        assert doc[1]._.lemmaW_ == ''  
        assert doc[2]._.lemmaW_ == 'laufen'          
        
        if verbose:
            print('wortschatz Tests 1 OK')          
            
            
        # Test2: doc erstellen
        text = self.TESTTEXTE[2]
        sent_anz = text.count('\n')+1        
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz  
        
        if verbose:
            print('wortschatz Tests 2 OK')   
            
            
        # Test3: doc erstellen
        text = self.TESTTEXTE[3]
        sent_anz = text.count('.')       
        text = self.blp.preprocess(text)
        doc = self.blp.nlp(text)
        
        # sent
        assert len(list(doc.sents)) == sent_anz          
          
        if verbose:
            print('wortschatz Tests 3 OK')    
            
            
        # Test4: Mehr oder weniger leeres Dokument
        texte = self.sent_start + self.sent_end + ['']
        for text in texte:
            #print(text)
            doc = self.blp.nlp(text)
            assert len(doc) == len(text)     
            
            doc = self.blp.nlp(text + '.')
            assert len(doc) <= len(text) + 1     
            
            doc = self.blp.nlp('»' + text)
            assert len(doc) == len(text) + 1            
            
        
        if verbose:
            print('wortschatz Tests 4 OK')           
            
        return True
      
            
            
#          # Wortschatz prüfen    
#          prüfe_tagZ( 'X_konj',        ['z.B.','z. B.','Z.B.', 'D.h.','D. h.','d.h.', 'd. h.']   , doc )
#          prüfe_tagZ( 'ADV',           ['U.a.','u.a.','u. a.']   , doc )            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            



#print('bj_nlp init')


   
# ==================================================================================================
# bj_nlp Files
# ==================================================================================================

from bj_nlp.Testtext         import *
from bj_nlp.Preprocessing    import *
from bj_nlp.blpClass         import *
from bj_nlp.translate_tags   import *
from bj_nlp._PandaziseMixin  import *
from bj_nlp._UtilsMixin      import *


# Custom Components. 
# Alle Custom Components sind in blpClass registriert.
from bj_nlp.c01_merger       import *
from bj_nlp.c05_wrede        import *
from bj_nlp.c10_wortschatz   import *








